/*
 * This module installs a NMI handler that is accessed via a task gate.
 * It is used to verify two things:
 * - virtual-NMI blocking state when task switch VM-exit happens due to NMI
 * - virtual-NMI blocking state when task switch VM-exit happens due to iret
 *
 * XXX this module works only when an NMI is injected on one cpu at a time.
 * This is because the TSS used by the NMI handler is shared by the vcpus.
 */
#include <sys/param.h>
#include <sys/kernel.h>
#include <sys/module.h>
#include <sys/sysctl.h>
#include <sys/systm.h>
#include <sys/bus.h>
#include <sys/pcpu.h>

#include <machine/pcb.h>
#include <machine/intr_machdep.h>
#include <machine/segments.h>
#include <machine/tss.h>

#ifndef __i386__
#error "i386 only module"
#endif

extern void dblfault_handler(void);
extern inthand_t IDTVEC(tsnmi);
extern void tsnmi_handler(void);

static SYSCTL_NODE(_debug, OID_AUTO, tsnmi, CTLFLAG_RW, 0, "");

static int num_nmi;
SYSCTL_INT(_debug_tsnmi, OID_AUTO, num, CTLFLAG_RW, &num_nmi, 0, NULL);

static struct gate_descriptor nmigate;

#if 0
static void
dumptss(struct i386tss *tss)
{
	printf("link: 0x%08x\n", tss->tss_link);
	printf("esp0: 0x%08x\n", tss->tss_esp0);
	printf("SS0:  0x%08x\n", tss->tss_ss0);
	printf("esp1: 0x%08x\n", tss->tss_esp1);
	printf("SS1:  0x%08x\n", tss->tss_ss1);
	printf("esp2: 0x%08x\n", tss->tss_esp2);
	printf("SS2:  0x%08x\n", tss->tss_ss2);
	printf("CR3:  0x%08x\n", tss->tss_cr3);
	printf("eip:  0x%08x\n", tss->tss_eip);
	printf("eflags: 0x%08x\n", tss->tss_eflags);
	printf("esp:  0x%08x\n", tss->tss_esp);
	printf("ES:  0x%08x\n", tss->tss_es);
	printf("CS:  0x%08x\n", tss->tss_cs);
	printf("SS:  0x%08x\n", tss->tss_ss);
	printf("DS:  0x%08x\n", tss->tss_ds);
	printf("FS:  0x%08x\n", tss->tss_fs);
	printf("GS:  0x%08x\n", tss->tss_gs);
	printf("LDT:  0x%08x\n", tss->tss_ldt);
	printf("ioopt:  0x%08x\n", tss->tss_ioopt);
}
#endif

extern struct pcpu __pcpu[];

void
tsnmi_handler(void)
{
	struct i386tss *curtss, *prevtss;
	struct segment_descriptor *sd;
	int eflags;

	atomic_add_int(&num_nmi, 1);

	eflags = read_eflags();
	KASSERT((eflags & PSL_I) == 0, ("interrupts must be disabled"));
	KASSERT((eflags & PSL_NT) != 0, ("nested task must be enabled"));

	sd = &gdt[curcpu * NGDT + GPANIC_SEL].sd;
	curtss = (struct i386tss *)(uintptr_t)USD_GETBASE(sd);

	sd = &gdt[curcpu * NGDT + IDXSEL(curtss->tss_link)].sd;
	prevtss = (struct i386tss *)(uintptr_t)USD_GETBASE(sd);
	prevtss->tss_cr3 = curpcb->pcb_cr3;
}

static void
install_tsnmi_handler(void)
{
	struct segment_descriptor *sd;
	struct i386tss *tss;

	sd = &gdt[GPANIC_SEL].sd;	/* double fault TSS descriptor */
	tss = (struct i386tss *)(uintptr_t)USD_GETBASE(sd);
	tss->tss_eip = (int)IDTVEC(tsnmi);

	/*
	 * Save original NMI gate descriptor and install a task gate in
	 * its place.
	 */
	nmigate = idt[IDT_NMI];
	setidt(IDT_NMI, 0,  SDT_SYSTASKGT, SEL_KPL, GSEL(GPANIC_SEL, SEL_KPL));
}

static void
uninstall_tsnmi_handler(void)
{
	struct segment_descriptor *sd;
	struct i386tss *tss;

	/* restore original NMI handler */
	idt[IDT_NMI] = nmigate;

	/* restore dblfault handler */
	sd = &gdt[GPANIC_SEL].sd;
	tss = (struct i386tss *)(uintptr_t)USD_GETBASE(sd);
	tss->tss_eip = (int)dblfault_handler;
}

static int
tsnmi_modevent(module_t mod, int type, void *unused)
{
        switch (type) {
        case MOD_LOAD:
		install_tsnmi_handler();
		break;
	case MOD_UNLOAD:
		uninstall_tsnmi_handler();
		break;
	default:
		return (EINVAL);
	}

	return (0);
}

static moduledata_t tsnmi_mod_data= {
	"tsnmi",
	tsnmi_modevent,
};

DECLARE_MODULE(test, tsnmi_mod_data, SI_SUB_SMP + 1, SI_ORDER_ANY);
